@@ -1,7 +1,7 @@
 package JobCenter::Client::Mojo;
 use Mojo::Base -base;
 
-our $VERSION = '0.19'; # VERSION
+our $VERSION = '0.20'; # VERSION
 
 #
 # Mojo's default reactor uses EV, and EV does not play nice with signals
@@ -13,14 +13,19 @@ BEGIN {
 }
 # more Mojolicious
 use Mojo::IOLoop;
+use Mojo::IOLoop::Stream;
 use Mojo::Log;
 
 # standard perl
 use Carp qw(croak);
 use Cwd qw(realpath);
+use Data::Dumper;
 use Encode qw(encode_utf8 decode_utf8);
 use File::Basename;
 use FindBin;
+use IO::Handle;
+use POSIX ();
+use Storable;
 use Sys::Hostname;
 
 # from cpan
@@ -28,7 +33,7 @@ use JSON::RPC2::TwoWay;
 # JSON::RPC2::TwoWay depends on JSON::MaybeXS anyways, so it can be used here
 # without adding another dependency
 use JSON::MaybeXS qw(decode_json encode_json);
-use MojoX::NetstringStream 0.04;
+use MojoX::NetstringStream 0.04; # older versions have utf-8 bugs
 
 has [qw(
 	actions address auth conn daemon debug jobs json log method 
@@ -340,7 +345,10 @@ sub announce {
 	my ($self, %args) = @_;
 	my $actionname = $args{actionname} or croak 'no actionname?';
 	my $cb = $args{cb} or croak 'no cb?';
-	my $async = $args{async} // 0;
+	#my $async = $args{async} // 0;
+	my $mode = $args{mode} // (($args{async}) ? 'async' : 'sync');
+	croak "unknown callback mode $mode" unless $mode =~ /^(subproc|async|sync)$/;
+	my $undocb = $args{undocb};
 	my $slots = $args{slots} // 1;
 	my $host = hostname;
 	my $workername = $args{workername} // "$self->{who} $host $0 $$";
@@ -367,7 +375,11 @@ sub announce {
 		}
 		my ($res, $msg) = @$r;
 		$self->log->debug("announce got res: $res msg: $msg");
-		$self->actions->{$actionname} = { cb => $cb, async => $async, slots => $slots } if $res;
+		$self->actions->{$actionname} = {
+			cb => $cb,
+			mode => $mode,
+			undocb => $undocb,
+			slots => $slots } if $res;
 		$err = $msg unless $res;
 	})->catch(sub {
 		my $d;
@@ -416,21 +428,102 @@ sub rpc_task_ready {
 			return;
 		}
 		local $@;
-		if ($action->{async}) {
+		if ($action->{mode} eq 'subproc') {
+			eval {
+				$self->_subproc($c, $action, $job_id, $cookie, $inargs);
+			};
+			$c->notify('task_done', { cookie => $cookie, outargs => { error => $@ } }) if $@;
+		} elsif ($action->{mode} eq 'async') {
 			eval {
 				$action->{cb}->($job_id, $inargs, sub {
 					$c->notify('task_done', { cookie => $cookie, outargs => $_[0] });
 				});
 			};
 			$c->notify('task_done', { cookie => $cookie, outargs => { error => $@ } }) if $@;
-		} else { 
+		} elsif ($action->{mode} eq 'sync') {
 			my $outargs = eval { $action->{cb}->($job_id, $inargs) };
 			$outargs = { error => $@ } if $@;
 			$c->notify('task_done', { cookie => $cookie, outargs => $outargs });
+		} else { 
+			die "unkown mode $action->{mode}";
 		}
 	});
 }
 
+sub _subproc {
+	my ($self, $c, $action, $job_id, $cookie, $inargs) = @_;
+
+	# based on Mojo::IOLoop::Subprocess
+	my $ioloop = Mojo::IOLoop->singleton;
+
+	# Pipe for subprocess communication
+	pipe(my $reader, my $writer) or die "Can't create pipe: $!";
+
+	die "Can't fork: $!" unless defined(my $pid = fork);
+	unless ($pid) {# Child
+		$self->log->debug("in child $$");;
+		$ioloop->reset;
+		close $reader; # or we won't get a sigpipe when daddy dies..
+		my $undo = 0;
+		my $outargs = eval { $action->{cb}->($job_id, $inargs) };
+		if ($@) {
+			$outargs = {'error' => $@};
+			$undo++;
+		} elsif (ref $outargs eq 'HASH' and $outargs->{'error'}) {
+			$undo++;
+		}
+		if ($undo and $action->{undocb}) {
+			$self->log->info("undoing for $job_id");;
+			my $res = eval { $action->{undocb}->($job_id, $inargs); };
+			$res = $@ if $@;
+			# how should this look?
+			$outargs = {'error' => {
+				'msg' => 'undo failure',
+				'undo' => $res,
+				'olderr' => $outargs->{error},
+			}};
+			$undo = 0;
+		}
+		# stop ignoring sigpipe
+		$SIG{PIPE} = sub { $undo++ };
+		# if the parent is gone we get a sigpipe here:
+		print $writer Storable::freeze($outargs);
+		$writer->flush or $undo++;
+		close $writer or $undo++;
+		if ($undo and $action->{undocb}) {
+			$self->log->info("undoing for $job_id");;
+			eval { $action->{undocb}->($job_id, $inargs); };
+			# ignore errors because we can't report them back..
+		}
+		# FIXME: normal exit?
+		POSIX::_exit(0);
+	}
+
+	# Parent
+	my $me = $$;
+	close $writer;
+	my $stream = Mojo::IOLoop::Stream->new($reader)->timeout(0);
+	$ioloop->stream($stream);
+	my $buffer = '';
+	$stream->on(read => sub { $buffer .= pop });
+	$stream->on(
+		close => sub {
+			#say "close handler!";
+			return unless $$ == $me;
+			waitpid $pid, 0;
+			my $outargs = eval { Storable::thaw($buffer) };
+			$outargs = { error => $@ } if $@;
+			if ($outargs and ref $outargs eq 'HASH') {
+				$self->log->debug('subprocess results: ' . Dumper($outargs));
+				eval { $c->notify(
+						'task_done',
+						{ cookie => $cookie, outargs => $outargs }
+				); }; # the connection might be gone?
+			} # else?
+		}
+	);
+}
+
 # copied from Mojo::Server
 sub _daemonize {
 	use POSIX;
@@ -642,8 +735,29 @@ Valid arguments are:
 
 (required)
 
-=item - async: if true then the callback gets passed another callback as the
-last argument that is to be called on completion of the task.
+=item - mode: callback mode
+
+(optional, default 'sync')
+
+Possible values:
+
+=over 8
+
+=item - 'sync': simple blocking mode, just return the results from the
+callback.  Use only for callbacks taking less than (about) a second.
+
+=item - 'subproc': the simple blocking callback is started in a seperate
+process.  Useful for callbacks that take a long time.
+
+=item - 'async': the callback gets passed another callback as the last
+argument that is to be called on completion of the task.  For advanced use
+cases where the worker is actually more like a proxy.  The (initial)
+callback is expected to return soonish to the event loop, after setting up
+some Mojo-callbacks.
+
+=back
+
+=item - async: backwards compatible way for specifying mode 'async'
 
 (optional, default false)
 
@@ -652,6 +766,12 @@ for this action.
 
 (optional, default 1)
 
+=item - undocb: a callback that gets called when the original callback
+returns an error object or throws an error.  Called with the same arguments
+as the original callback.
+
+(optional, only valid for mode 'subproc')
+
 =back
 
 =head2 work
